module HuggORM
  module Persistence
    include Comparable

    def self.included(base)
      base.send(:include, InstanceMethods)
      base.extend(ClassMethods)
    end

    module ClassMethods
      attr_accessor :_table_name, :_primary_key, :_key_field, :_exclude_fields, :_include_fields

      def table_name(table_name = nil)
        self._table_name = table_name unless table_name.nil?
        self._table_name
      end

      def primary_key(primary_key = nil)
        self._primary_key = primary_key unless primary_key.nil?
        self._primary_key ||= "data->>'#{key_field}'"
      end

      def key_field(key_field = nil)
        self._key_field = key_field.to_sym unless key_field.nil?
        self._key_field ||= :id
      end

      def exclude_fields(*fields)
        self._exclude_fields = fields unless fields.empty?
        self._exclude_fields ||= []
      end

      def include_fields(*fields)
        self._include_fields = fields unless fields.empty?
        self._include_fields ||= []
      end

      def find(id)
        result = where("#{primary_key} = ?", [id]).select.first
        new(result) if result
      end

      def find!(id)
        find(id) || raise(RecordNotFound.new(id))
      end

      def all
        results = new_query.select
        new_collection(results)
      end

      def delete_all
        new_query.delete
      end

      def select(query = nil, where: nil, order: nil, limit: nil, offset: nil)
        return new_collection(query.select) if query

        query = new_query
        query.where(where[:conditions], where[:values]) if where
        query.order(order) if order
        query.limit(limit) if limit
        query.offset(offset) if offset

        new_collection(query.select)
      end

      def create(data)
        new(data) if new_query.insert(data) > 0
      end

      def update(id, data)
        results = where("#{primary_key} = ?", [id]).update(data, true)
        new(results.first) unless results.empty?
      end

      def update!(id, data)
        update(id, data) || raise(RecordNotFound.new(id))
      end

      def destroy(id)
        where("#{primary_key} = ?", [id]).delete > 0
      end

      def destroy!(id)
        destroy(id) || raise(RecordNotFound.new(id))
      end

      def count
        new_query.count
      end

      def where(conditions, values = [])
        new_query.where(conditions, values)
      end

      def limit(limit, offset = nil)
        new_query.limit(limit, offset)
      end

      def order(*args)
        new_query.order(*args)
      end

      def new_query
        Query.new(table_name)
      end

      def new_collection(data_ary)
        data_ary.map { |data| new(data) }
      end

      def to_pgtimestamp(field)
        "to_timestamp(#{field}, 'YYYY-MM-DD HH24:MI:SS')"
      end

      def string_to_pgtimestamp(string)
        to_pgtimestamp("'#{string}'")
      end

      def escape_like(str = '')
        pat = Regexp.union("\\", "%", "_")
        str.gsub(pat) { |mat| ["\\", mat].join }
      end

    end

    module InstanceMethods
      attr_accessor :uuid, :data

      def initialize(data = {})
        @data = {}
        update_attributes(data) unless data.empty?
      end

      def key
        @data[self.class.key_field]
      end

      def set_id
        # Set id based on key field, but only if id isn't already set and
        # the key method has been overridden and returns data.
        @data[self.class.key_field] = key if @data[self.class.key_field].nil? && key
      end

      def update_attributes(update_data)
        # Remove fields not set to be included
        unless self.class.include_fields.empty?
          update_data = self.class.include_fields.each_with_object({}) do |key, hash|
            hash[key] = update_data[key] if update_data.has_key? key
          end
        end

        # Update all fields unless they have been excluded
        update_data.each do |field, value|
          @data[field] = value unless self.class.exclude_fields.include?(field)
        end

        # Set id based on key method.
        set_id

        self
      end

      def destroy
        if self.class.destroy(key)
          @data = {}
          @state = :destroyed
        end

        self
      end

      def save
        updated_data = self.class.where("#{self.class.primary_key} = ?", [key]).update(@data, true)

        unless updated_data.empty?
          @data = updated_data.first
          @state = :updated
        else
          self.class.new_query.insert(@data)
          @state = :created
        end

        self
      end

      def destroyed?
        @state == :destroyed
      end

      def updated?
        @state == :updated
      end

      def created?
        @state == :created
      end

      def dump
        str = "#{self.class.name}:\n"
        @data.sort.each do |field, value|
          str << "  #{field}: #{value}\n"
        end
        str
      end

      def fetch(*args)
        @data.send(:fetch, *args)
      end

      def has_field?(field)
        @data.has_key?(field)
      end

      def <=>(o)
        o.kind_of?(self.class) && @data == o.data ? 0 : nil
      end
    end
  end

  class RecordNotFound < StandardError; end
end
